<?php

use Myriad\Installations\Models\Installation;
use Behat\Gherkin\Node\PyStringNode;
use Behat\Behat\Tester\Exception\PendingException;
use Myriad\Authentication\Models\User;
use Behat\Testwork\Hook\Scope\BeforeSuiteScope;
use Behat\Testwork\Hook\Scope\AfterSuiteScope;
use Behat\Behat\Context\SnippetAcceptingContext;
use Behat\Behat\Hook\Scope\BeforeScenarioScope;
use Illuminate\Support\Facades\Hash;
use Myriad\Authentication\Models\UserEmail;
use Ramsey\Uuid\Uuid;
use Myriad\Authentication\Lib\Repositories\EloquentAccessTokenRepository;
use Illuminate\Http\Request;
use Illuminate\Foundation\Http\Kernel as HttpKernel;
use Illuminate\Support\Facades\Auth;
use GuzzleHttp\Client;
use Myriad\Authentication\Models\VerificationCode;
use Carbon\Carbon;
use Myriad\Authentication\Lib\Repositories\EloquentVerificationCodeRepository;
use Behat\Behat\Hook\Scope\AfterStepScope;
use Behat\Behat\Hook\Scope\BeforeStepScope;
use Myriad\Authentication\Lib\Strategies\EmailVerification;

class FeatureAuthenticate
{
  /**
   * Handle an incoming request.
   *
   * @param  \Illuminate\Http\Request  $request
   * @param  \Closure  $next
   * @param  string|null  $guard
   * @return mixed
   */
  public function handle($request, Closure $next, $guard = null)
  {
    if (Auth::guard($guard)->guest()) {
      if ($request->ajax() || $request->wantsJson()) {
        return response('Unauthorized.', 401);
      } else {
        return redirect()->guest('login');
      }
    }

    return $next($request);
  }
}


class FeatureKernel extends HttpKernel {
  /**
   * The application's route middleware.
   *
   * These middleware may be assigned to groups or used individually.
   *
   * @var array
   */
  protected $routeMiddleware = [
    'auth' => FeatureAuthenticate::class
  ];
}

class ApiContext implements SnippetAcceptingContext {

  use BootstrapsKernel;
  use RunsMigrations;
  use InteractsWithMailcatcher;
  private static $settings;

  const LOGIN_URL = 'api/v1/auth/login';
  const ME_URL = 'api/v1/auth/me';
  const REGISTER_URL = 'api/v1/auth/register';
  const VERIFY_CODE_URL = 'api/v1/auth/verification_codes/verify';
  const SEND_CODE_URL = 'api/v1/auth/verification_codes/send';

  private $guzzleClient;

  private $accessTokenRepository;
  private $verificationCodeRepository;
  private $verificationCode;
  private $firstName;
  private $lastName;
  private $password = 'password12345';
  private $email = 'somebody@example.com';
  private $phoneNumber = '5005550006';
  private $user;
  private $installation;
  private $accessToken;
  private $verifyCodeResponse;


  public function __construct($baseUrl, $mailcatcherBaseUrl) {
    $this->accessTokenRepository = self::$app->make(EloquentAccessTokenRepository::class);
    $this->verificationCodeRepository = self::$app->make(EloquentVerificationCodeRepository::class);

    self::$mailcatcherGuzzleClient = new Client([
      'base_url' => $mailcatcherBaseUrl
    ]);
  }

  /**
   * @BeforeSuite
   */
  public static function beforeSuite(BeforeSuiteScope $scope) {
    self::$settings = $scope->getSuite()->getSettings()['parameters'];

    self::bootstrapKernel(self::$settings, 'FeatureKernel');
  }

  /**
   * @BeforeScenario
   */
  public static function BeforeScenario(BeforeScenarioScope $scope) {
    self::runMigrations();
    self::deleteAllMailcatcherEmails();
  }

  /**
   * @Given my installation already exists in the system
   */
  public function myInstallationAlreadyExistsInTheSystem()
  {
    $this->installation = new Installation([
      'app_name' => 'test_app',
      'app_version' => '12345'
    ]);
    $this->installation->uuid = Uuid::uuid4()->toString();
    $this->installation->save();
  }

  /**
   * @Given I already registered with the system via email
   */
  public function iAlreadyRegisteredWithTheSystemViaEmail()
  {
    $this->user = User::create([
      "first_name" => $this->firstName = "george",
      "last_name" => $this->lastName = "washington",
      'password' => $this->password
    ]);

    $this->userEmail = $this->user->emailAddresses()->create([
      'email' => $this->email,
      'primary' => true
    ]);


    $verificationCode = VerificationCode::create([
      'contact' => $this->email,
      'code' => '12345',
      'verification_strategy_type' => EmailVerification::class,
      'data' => json_encode([
        'user_id' => $this->user->id,
        'type' => 'personal',
        'primary' => true
      ]),
      'expires_at' => Carbon::now()->addMinutes(1)
    ]);

    $this->verificationCode = $verificationCode->code;

    $this->accessToken = $this->accessTokenRepository->createAccessToken($this->user, $this->installation)->token;
  }

  /**
   * @When I login with using valid email and password
   */
  public function iLoginWithUsingValidEmailAndPassword()
  {
    $request = Request::create(self::LOGIN_URL, 'POST', [
      'email' => $this->email,
      'password' => $this->password,
    ], [], [], $this->headers());

    $this->response = self::$kernel->handle($request);
    self::$kernel->terminate($request, $this->response);
  }

  /**
   * @Given there are no users in the system
   */
  public function thereAreNoUsersInTheSystem()
  {
    User::getQuery()->delete();
  }

  /**
   * @When I submit a valid registration
   */
  public function iSubmitAValidRegistration()
  {
    $request = Request::create(self::REGISTER_URL, 'POST', [
      "first_name" => "george",
      "last_name" => "washington",
      "email" => $this->email,
      "type" => "personal",
      "password" => $this->password,
      "password_confirmation" => $this->password
    ], [], [], $this->headers());

    $this->response = self::$kernel->handle($request);
    self::$kernel->terminate($request, $this->response);
  }

  private function headers($auth = false) {
    $headers = [
      "HTTP_App-Name" => "test-app",
      "HTTP_App-Version" => "12345-ios",
      "HTTP_App-Installation-Id" => $this->installation->uuid,
      "HTTP_User-Agent" => "ios-device"
    ];

    if ($auth) {
      $headers['HTTP_Authorization'] = "Token " . $this->accessToken;
    }

    return $headers;
  }

  /**
   * @Then the response should contain an access token and its expiration
   */
  public function theResponseShouldContainAnAccessTokenAndItsExpiration()
  {
    $json = json_decode($this->response->getContent(), true);
    PHPUnit_Framework_Assert::assertArrayHasKey('data', $json);
    PHPUnit_Framework_Assert::assertArrayHasKey('access_token', $json['data']);
    PHPUnit_Framework_Assert::assertArrayHasKey('expires_at', $json['data']);
  }

  /**
   * @When I request my information from the system
   */
  public function iRequestMyInformationFromTheSystem()
  {
    $request = Request::create(self::ME_URL, 'GET', [], [], [], $this->headers(true));

    $this->response = self::$kernel->handle($request);

    self::$kernel->terminate($request, $this->response);
  }

  /**
   * @Then my information should contain my basic information
   */
  public function myInformationShouldContainMyBasicInformation()
  {
    $content = json_decode($this->response->getContent(), true);
    \PHPUnit_Framework_Assert::assertArrayHasKey('first_name', $content['data']);
    \PHPUnit_Framework_Assert::assertArrayHasKey('last_name', $content['data']);
    \PHPUnit_Framework_Assert::assertArrayHasKey('last_login', $content['data']);
  }

  /**
   * @Then my information should contain my email address
   */
  public function myInformationShouldContainMyEmailAddress()
  {
    $content = json_decode($this->response->getContent(), true);
    \PHPUnit_Framework_Assert::assertCount(1, $content['data']['email_addresses']['data']);
    \PHPUnit_Framework_Assert::assertEquals($this->email, $content['data']['email_addresses']['data'][0]['email']);
  }

  /**
   * @Then my information should contain an empty list of phone numbers
   */
  public function myInformationShouldContainAnEmptyListOfPhoneNumbers()
  {
    $content = json_decode($this->response->getContent(), true);
    \PHPUnit_Framework_Assert::assertCount(0, $content['data']['phone_numbers']['data']);
  }

  /**
   * @Given I received an email with an email auth code
   */
  public function iReceivedAnEmailWithAnEmailAuthCode()
  {
    //$this->verificationAuthCode = $this->verificationCode->code;
    throw new PendingException();
  }

  /**
   * @When I verify my auth code
   */
  public function iVerifyMyAuthCode()
  {
    $request = Request::create(self::VERIFY_CODE_URL, 'POST', [
      'contact' => $this->email,
      'code' => $this->verificationCode
    ], [], [], $this->headers());

    $this->verifyCodeResponse = self::$kernel->handle($request);

    self::$kernel->terminate($request, $this->verifyCodeResponse);

    \PHPUnit_Framework_Assert::assertEquals(200, $this->verifyCodeResponse->getStatusCode());

  }

  /**
   * @Then my information should contain my verified email address
   */
  public function myInformationShouldContainMyVerifiedEmailAddress()
  {
    $content = json_decode($this->response->getContent(), true);
    \PHPUnit_Framework_Assert::assertCount(1, $content['data']['email_addresses']['data']);

    $email = $content['data']['email_addresses']['data'][0];
    \PHPUnit_Framework_Assert::assertEquals($this->email, $email['email'] );

    \PHPUnit_Framework_Assert::assertEquals(true, $email['verified']);
  }

  /**
   * @When I request an verification code to be sent to my email address
   */
  public function iRequestAnVerificationCodeToBeSentToMyEmailAddress()
  {
    $request = Request::create(self::SEND_CODE_URL, 'POST', [
      'email' => $this->email
    ], [], [], $this->headers());

    $this->response = self::$kernel->handle($request);
    self::$kernel->terminate($request, $this->response);
  }

  /**
   * @Then I should receive an email containing the verification code
   */
  public function iShouldReceiveAnEmailContainingTheVerificationCode()
  {
    $mailcatcherEmails = self::getAllMailcatcherEmails();
    \PHPUnit_Framework_Assert::assertCount(1, $mailcatcherEmails);
  }

  /**
   * @When I request an verification code to be sent to my phone
   */
  public function iRequestAnVerificationCodeToBeSentToMyPhone()
  {
    $request = Request::create(self::SEND_CODE_URL, 'POST', [
      'phone' => $this->phoneNumber
    ], [], [], $this->headers());

    $this->response = self::$kernel->handle($request);
    self::$kernel->terminate($request, $this->response);
  }

  /**
   * @Then I should receive an text message containing the verification code
   */
  public function iShouldReceiveAnTextMessageContainingTheVerificationCode()
  {
    //TODO: figure out how to test this
  }

  /**
   * @When I obtain the auth code from my email
   */
  public function iObtainTheAuthCodeFromMyEmail()
  {
    $emails = $this->getAllMailcatcherEmails();
    $email = $this->getMailcatcherEmail($emails[0]["id"]);
    $message = $email["source"];

    $pattern = '/enter verification code (.{4})/';

    preg_match($pattern, $message, $matches);

    $code = $matches[1];

    $this->verificationCode = $code;
  }
}
