<?php

namespace Myriad\Authentication;

use Illuminate\Contracts\Foundation\Application;
use Illuminate\Support\ServiceProvider;
use Illuminate\Database\Eloquent\Relations\Relation;
use Myriad\Authentication\Lib\Repositories\IAccessTokenRepository;
use Myriad\Authentication\Lib\Repositories\EloquentAccessTokenRepository;
use Illuminate\Auth\AuthManager;
use Myriad\Authentication\Lib\Providers\EloquentUserProvider;
use Illuminate\Contracts\Hashing\Hasher;
use Myriad\Authentication\Lib\Guards\AccessTokenGuard;
use Myriad\Authentication\Http\Transformers\IUserTransformer;
use Illuminate\Contracts\Validation\Factory as ValidationFactory;
use Myriad\Authentication\Lib\Repositories\IVerificationCodeRepository;
use Myriad\Authentication\Lib\Repositories\EloquentVerificationCodeRepository;
use Myriad\Authentication\Lib\Repositories\IUserRepository;
use Myriad\Authentication\Lib\Listeners\UserRegisteredWithEmailListener;
use Myriad\Authentication\Lib\Services\IEmailConfirmer;
use Myriad\Authentication\Lib\Strategies\EmailVerification;
use Myriad\Authentication\Lib\Services\IPhoneNumberConfirmer;
use Myriad\Authentication\Lib\Strategies\PhoneNumberVerification;
use Myriad\Authentication\Lib\Services\ITextMessageSender;
use Myriad\Authentication\Lib\Services\MockTextMessageSender;

/**
 * Class: AuthenticationServiceProvider
 *
 * @see ServiceProvider
 */
class AuthenticationServiceProvider extends ServiceProvider {

  /**
   * Perform post-registration booting of services.
   *
   * @return void
   */
  public function boot(
    Application $app,
    AuthManager $auth,
    ValidationFactory $validationFactory
  ) {
    //TODO: fix this...
    //      capsule seems to be overriding fetch and default
    //      connection and i have no idea why. quickfix is to cache existing settings
    //      and then reset them after creating capsule manager
    $databaseFetch = $app['config']['database.fetch'];
    $defaultConnection = $app['config']['database.default'];
    $capsule = $app->make('Illuminate\Database\Capsule\Manager');
    $app['config']['database.fetch'] = $databaseFetch;
    $app['config']['database.default'] = $defaultConnection;
    $connection = $app['config']['database.connections.'. $app['config']['database.default']];
    $capsule->addConnection($connection);
    $capsule->setAsGlobal();

    // include routes
    if (!$this->app->routesAreCached()) {
      require __DIR__ . '/routes.php';
    }

    // publish config and migrations
    $this->publishes([__DIR__ . '/../database/migrations/' => database_path('migrations') ], 'migrations');
    $this->publishes([__DIR__ . '/../config/authentication.php' => config_path('authentication.php')], 'config');

    // configure morphMap
    $morphMap = config('authentication.morph_map');

    if (!is_null($morphMap)) {
      Relation::morphMap($morphMap);
    }

    $auth->provider('myriad_user_provider', function($app, array $config) {
      return new EloquentUserProvider(
        $app->make(Hasher::class),
        $app->make(IAccessTokenRepository::class),
        $config['model']
      );
    });

    $auth->extend('myriad_access_token_guard', function($app, $name, array $config) use ($auth) {
      return new AccessTokenGuard($auth->createUserProvider($config['provider']), $app['request']);
    });

    $this->loadViewsFrom(__DIR__.'/../resources/views', 'authentication');
    $this->loadTranslationsFrom(__DIR__.'/../resources/lang', 'unique_verified');

    $validationFactory->extend('unique_verified', function($attribute, $value, $parameters, $validator) {
      $parameters = array_map('trim', $parameters);
      $table = $parameters[0];
      $column = $attribute;

      $verifier = $validator->getPresenceVerifier();
      $count = $verifier->getCount(
        $table,
        $column,
        $value,
        null,
        null,
        [
          'verified' => true
        ]
      );

      return $count === 0;
    });
  }

  /**
   * Register the service provider.
   *
   * @return void
   */
  public function register() {
    $this->app->bind(IAccessTokenRepository::class, EloquentAccessTokenRepository::class);
    $this->app->bind(IUserTransformer::class, config('authentication.user_transformer'));
    $this->app->bind(IUserRepository::class, config('authentication.user_repository'));
    $this->app->bind(IVerificationCodeRepository::class, EloquentVerificationCodeRepository::class);

    $this->app->bind(IEmailConfirmer::class, config('authentication.verification_strategies.email', EmailVerification::class));
    $this->app->bind(IPhoneNumberConfirmer::class, config('authentication.verification_strategies.phone', PhoneNumberVerification::class));

    if (config('app.env') === 'testing') {
      $this->app->bind(ITextMessageSender::class, MockTextMessageSender::class);
    } else {
      $this->app->bind(ITextMessageSender::class, config('authentication.text_messages.text_message_sender.class'));
    }
  }

}
