<?php

namespace Myriad\Authentication\Lib\Strategies;


trait VerifiesUserContactMethod {

  protected $userRepository;
  protected $model;
  protected $property;
  protected $contactCollection;
  protected $contactAlreadyExistsError;

  /**
   * Perform post-verification strategy.
   *
   * This method looks at the data associated with the token to see how to
   * handle syncing up the code with a user account.
   *
   * Scenarios:
   * + Verification code was created while user was logged in:
   *   + If contact (email/phone) in code belongs to different user
   *     then throw exception
   *   + If contact belongs to current user then create contact record
   *     if it doesn't exist or update exising if it does
   * + Verification code was created with non-logged in user:
   *   + If contact method belongs to existing user then set that user as
   *     the current user
   *   + If no user exists for that contact, then create a new user and
   *     create a new contact association.
   *
   * @param string $code - The code being to validate against code in $data
   * @param array $data - The data associated with the Verification Code
   *
   * @return \Illuminate\Contracts\Auth\Authenticatable
   */
  public function performVerificationStrategy($code, array $data) {
    if ($code !== $data['code']) {
      throw new \Exception('Invalid code!');
    }

    $contactCollection = $this->contactCollection;
    $customData = json_decode($data['data'], true);
    $contact = trim($data['contact']);

    $existingUserContact = $this->createModel()
      ->newQuery()
      ->where($this->property, $contact)
      ->where('verified', true)
      ->first();

    // TODO: refactor this into private methods so it's easier to consume as reader
 
    // check if verification code was created when user was already logged in
    if (array_key_exists('user_id', $customData) && !is_null($customData['user_id'])) {
      $user = $this->userRepository->retrieveById($customData['user_id']);

      if ($existingUserContact) {
        // there are already a user associated with the contact and it has been verified
        if ($user->id !== $existingUserContact->user->id) {
          // the existing user who has already verified the contact previously differs
          // from the user associated with this code. this should hopefully never occur
          //TODO: throw a better exception
          throw new \Exception($this->contactAlreadyExistsError);
        }
      } else {
        // no existing user attached to contact property, so we're ok to create/update contact for our user
        $userContact = $user->$contactCollection()->where($this->property, $contact)->first();

        if ($userContact) {
          // contact already exists for user. make sure it is marked as verified (should already be verified)
          $userContact->verified = true;
          $userContact->save();
        } else {
          // contact does not exist for user. create a new one
          $user->$contactCollection()->create(array_merge(
            [ $this->property => $contact ],
            $customData,
            [ 'verified' => true ]
          ));
        }
      }
    } else {
      if ($existingUserContact) {
        // there is already an existing user associated with the contact. since it has
        // already been verified that the caller owns this contact method, the caller
        // and the existing user should be the same user. so, log them in
        $user = $existingUserContact->user;
      } else {
        // no existing user attached to contact, so we're ok to create a user and attach contact
        $user = $this->userRepository->createUser();
        
        $user->$contactCollection()->create(array_merge(
          [ $this->property => $contact ],
          $customData,
          [ 'verified' => true ]
        ));
      }
    }

    return $user;
  }
  
  /**
   * Create a new instance of the model.
   *
   * @return \Illuminate\Database\Eloquent\Model
   */
  protected function createModel()
  {
      $class = '\\'.ltrim($this->model, '\\');

      return new $class;
  }
}
